"""
badread.py -- Wrapper for BadRead
"""

__all__ = (
    "BadReadAdapter",
    "ALL_POSSIBLE_BADREAD_MODELS",
    "patch_frontend_parser"
)

import argparse

from labw_utils.typing_importer import List, Final, Mapping, Any
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, LLRGInitializationException

ALL_POSSIBLE_BADREAD_MODELS = ("nanopore2018", "nanopore2020", "pacbio2016", "verybad", "verynice")
"""All possible badread model names"""


class BadReadAdapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper for BadRead

    Cmdline Specs::

        if self.model_name == "verybad":
            cmd = [
                llrg_executable_path, "simulate",
                "--reference", self._src_fasta_file_path,
                "--quantity", f"{self._depth}x",
                "--glitches", "1000,100,100",
                "--junk_reads", "5",
                "--random_reads", "5",
                "--chimeras", "10",
                "--identity", "75,90,8",
                "--start_adapter_seq", "",
                "--end_adapter_seq", "",
                *other_args
            ]
        elif self.model_name == "verynice":
            cmd = [
                llrg_executable_path, "simulate",
                "--reference", self._src_fasta_file_path,
                "--quantity", f"{self._depth}x",
                "--error_model", "random",
                "--qscore_model", "ideal",
                "--glitches", "0,0,0",
                "--junk_reads", "0",
                "--random_reads", "0",
                "--chimeras", "0",
                "--identity", "95,100,4",
                "--start_adapter_seq", "",
                "--end_adapter_seq", "",
                *other_args
            ]
        else:
            cmd = [
                llrg_executable_path, "simulate",
                "--reference", self._src_fasta_file_path,
                "--quantity", f"{self._depth}x",
                "--error_model", model_name,
                "--qscore_model", model_name,
                "--start_adapter_seq", "",
                "--end_adapter_seq", "",
                *other_args
            ]
    """

    @staticmethod
    def validate_params(model_name: str, **kwargs) -> Mapping[str, Any]:
        if model_name not in ALL_POSSIBLE_BADREAD_MODELS:
            raise LLRGInitializationException(f"Illegal model name {model_name}")
        return {}

    llrg_name: Final[str] = "badread"
    _require_integer_depth: Final[bool] = True
    _capture_stdout: Final[bool] = True

    def __init__(
            self,
            *,
            src_fasta_file_path: str,
            dst_fastq_file_prefix: str,
            is_trusted: bool,
            depth: int,
            llrg_executable_path: str,
            model_name: str,
            other_args: List[str]
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param model_name: Name of pre-defined models. See :py:attr:`ALL_POSSIBLE_BADREAD_MODELS`.
        :param is_trusted: Whether to skip input validation test.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            is_trusted=is_trusted,
            preserve_intermediate_files=None  # Not effective.
        )
        if not is_trusted:
            _ = BadReadAdapter.validate_params(model_name=model_name)
        cmd = [
            llrg_executable_path, "simulate",
            "--reference", self._src_fasta_file_path,
            "--quantity", f"{self._depth}x",
        ]
        if model_name == "verybad":
            cmd.extend([
                "--glitches", "1000,100,100",
                "--junk_reads", "5",
                "--random_reads", "5",
                "--chimeras", "10",
                "--identity", "75,90,8"
            ])
        elif model_name == "verynice":
            cmd.extend([
                "--error_model", "random",
                "--qscore_model", "ideal",
                "--glitches", "0,0,0",
                "--junk_reads", "0",
                "--random_reads", "0",
                "--chimeras", "0",
                "--identity", "95,100,4"
            ])
        else:
            cmd.extend([
                "--error_model", model_name,
                "--qscore_model", model_name,
            ])
        cmd.extend([
            "--start_adapter_seq", "",
            "--end_adapter_seq", "",
            *other_args
        ])
        self._cmd = cmd

    def _post_execution_hook(self):
        """
        This function is passed since badread pours read into stdout.
        """
        pass

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    @property
    def is_pair_end(self) -> bool:
        return False


def patch_frontend_parser(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser with ART arguments.
    """

    parser.add_argument(
        '-m',
        '--model_name',
        required=True,
        help="Badread model name",
        nargs='?',
        type=str,
        action='store',
        choices=ALL_POSSIBLE_BADREAD_MODELS
    )
    return parser



"""
pbsim2.py -- Wrapper of PBSIM2.
"""
__all__ = (
    "Pbsim2Adapter",
    "PBSIM2_DIST_DIR_PATH",
    "PBSIM2_ALL_POSSIBLE_MODELS",
    "patch_frontend_parser"
)

import argparse
import glob
import os

from labw_utils.typing_importer import List, Final, Mapping, Any
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, automerge, LLRGInitializationException

PBSIM2_DIST_DIR_PATH = os.path.join(os.path.dirname(__file__), "pbsim2_dist")
"""
Where PBSIM2 stores its model.
"""

PBSIM2_ALL_POSSIBLE_MODELS = [
    os.path.basename(os.path.splitext(filename)[0])
    for filename in glob.glob(os.path.join(PBSIM2_DIST_DIR_PATH, "*.model"))
]
"""
Possible PBSIM2 models.
"""


class Pbsim2Adapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper of PBSIM2.

    CMDline Spec::

        cmd = [
            llrg_executable_path,
            "--prefix", self._tmp_dir,
            "--_depth", str(self._depth),
            "--hmm_model", self.hmm_model,
            *other_args,
            self._src_fasta_file_path
        ]
    """
    llrg_name: Final[str] = "pbsim2"
    _require_integer_depth: Final[bool] = False
    _capture_stdout: Final[bool] = False

    @staticmethod
    def validate_params(
            hmm_model: str,
            **kwargs
    ) -> Mapping[str, Any]:
        if os.path.exists(hmm_model):
            pass
        elif os.path.exists(os.path.join(PBSIM2_DIST_DIR_PATH, f"{hmm_model}.model")):
            hmm_model = os.path.join(PBSIM2_DIST_DIR_PATH, f"{hmm_model}.model")
        else:
            raise LLRGInitializationException(f"HMM Model {hmm_model} cannot be resolved!")
        return {"hmm_model": hmm_model}

    def __init__(
            self,
            *,
            src_fasta_file_path: str,
            dst_fastq_file_prefix: str,
            depth: int,
            llrg_executable_path: str,
            is_trusted: bool,
            hmm_model: str,
            preserve_intermediate_files: bool,
            other_args: List[str]
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :param is_trusted: Whether to skip input validation test.
        :param hmm_model: Name or Path of HMM Model.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            is_trusted=is_trusted,
            preserve_intermediate_files=preserve_intermediate_files
        )

        if not is_trusted:
            validated_params = Pbsim2Adapter.validate_params(hmm_model=hmm_model)
            hmm_model = validated_params["hmm_model"]
        self._cmd = [
            llrg_executable_path,
            "--prefix", os.path.join(self._tmp_dir, "tmp"),
            "--depth", str(self._depth),
            "--hmm_model", hmm_model,
            *other_args,
            self._src_fasta_file_path
        ]

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    def _post_execution_hook(self):
        automerge(glob.glob(os.path.join(self._tmp_dir, "tmp_????.fastq")), self._dst_fastq_file_prefix + ".fq")

    @property
    def is_pair_end(self) -> bool:
        return False


def patch_frontend_parser(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser with ART arguments.
    """
    parser.add_argument(
        '-m',
        '--hmm_model',
        required=True,
        help="Basename or absolute path of HMM file",
        nargs='?',
        type=str,
        action='store',
        choices=PBSIM2_ALL_POSSIBLE_MODELS
    )
    parser = patch_frontend_argument_parser(parser, "--preserve_intermediate_files")
    return parser



"""
pbsim3.py -- Wrapper of PBSIM3.
"""
__all__ = (
    "Pbsim3Adapter",
    "PBSIM3_DIST_DIR_PATH",
    "PBSIM3_STRATEGY",
    "patch_frontend_parser"
)

import argparse
import enum
import glob
import os
import subprocess
import time
import uuid

from labw_utils.bioutils.datastructure.fasta_view import FastaViewFactory
from labw_utils.commonutils.lwio.safe_io import get_writer
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import Final, List, Mapping, Any, Optional
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.helper.llrg import enhanced_which
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, autocopy, automerge, LLRGInitializationException

_lh = get_logger(__name__)

try:
    import jinja2

    PACB_SUBREAD_XML_TEMPLATE = (
        jinja2.
        Environment(
            loader=jinja2.PackageLoader('yasim.llrg_adapter', 'templates'),
            autoescape=True
        ).
        get_template('pbsim_xml_template.xml')
    )
    """PBSIM3 Subread XML Template."""
except ImportError:
    _lh.warning("PBSIM3: Jinja2 failed to import. CCS generation will be BAM-based instead of XML-based.")
    jinja2 = None
    PACB_SUBREAD_XML_TEMPLATE = None

PBSIM3_DIST_DIR_PATH = os.path.join(os.path.dirname(__file__), "pbsim3_dist")
"""
Where pbsim3 stores its models
"""


class PBSIM3_STRATEGY(enum.Enum):
    wgs = enum.auto()
    trans = enum.auto()

    @classmethod
    def from_name(cls, in_name: str) -> 'PBSIM3_STRATEGY':
        for choices in cls:
            if choices.name == in_name:
                return choices

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.name


PBSIM3_STRATEGY.wgs.__doc__ = "WGS mode (as PBSIM2)"
PBSIM3_STRATEGY.trans.__doc__ = "TRANS mode"

PBSIM3_QSHMM_POSSIBLE_MODELS = [
    os.path.basename(os.path.splitext(filename.replace("QSHMM-", ""))[0])
    for filename in glob.glob(os.path.join(PBSIM3_DIST_DIR_PATH, "QSHMM-*.model"))
]

PBSIM3_ERRHMM_POSSIBLE_MODELS = [
    os.path.basename(os.path.splitext(filename.replace("ERRHMM-", ""))[0])
    for filename in glob.glob(os.path.join(PBSIM3_DIST_DIR_PATH, "ERRHMM-*.model"))
]


class Pbsim3Adapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper of PBSIM3.

    CMDline Spec::

        cmd = [
            llrg_executable_path,
            "--strategy", "trans",
            "--method", self.hmm_method,
            f"--{self.hmm_method}", self.hmm_model,
            "--prefix", os.path.join(self._tmp_dir, "tmp"),
            "--transcript" if self.strategy == "trans" else "--genome", self._input_path,
            "--pass-num", str(self._ccs_pass),
            *other_args
        ]
    """
    _ccs_pass: int
    _samtools_executable_path: Optional[str]
    _ccs_executable_path: Optional[str]
    _ccs_num_threads: Optional[int]
    _strategy: PBSIM3_STRATEGY

    _input_path: str
    """Path to input transcript.tsv or genome.fasta"""

    llrg_name: Final[str] = "pbsim3"
    _require_integer_depth: Final[bool] = False
    _capture_stdout: Final[bool] = False

    @staticmethod
    def validate_params(
            hmm_model: str,
            hmm_method: str,
            ccs_pass: int,
            ccs_executable_path: Optional[str],
            samtools_executable_path: Optional[str],
            **kwargs
    ) -> Mapping[str, Any]:
        possible_hmm_model_path = os.path.join(
            PBSIM3_DIST_DIR_PATH,
            f"{hmm_method.upper()}-{hmm_model}.model"
        )
        if os.path.exists(hmm_model):
            pass
        elif os.path.exists(possible_hmm_model_path):
            hmm_model = possible_hmm_model_path
        else:
            raise LLRGInitializationException(f"HMM Model {hmm_model} cannot be resolved!")
        if ccs_pass > 1:
            try:
                ccs_executable_path = enhanced_which(ccs_executable_path)
            except FileNotFoundError as e:
                raise LLRGInitializationException(f"PBCCS at {ccs_executable_path} not found!") from e
            try:
                samtools_executable_path = enhanced_which(samtools_executable_path)
            except FileNotFoundError as e:
                raise LLRGInitializationException(f"SAMTOOLS at {samtools_executable_path} not found!") from e
        else:
            if ccs_executable_path is not None:
                _lh.warning("PBSIM3: CCS Executable path ignored in CLR mode")
            if samtools_executable_path is not None:
                _lh.warning("PBSIM3: SAMTOOLS Executable path ignored in CLR mode")
        return {
            "hmm_model": hmm_model,
            "ccs_executable_path": ccs_executable_path,
            "samtools_executable_path": samtools_executable_path
        }

    def __init__(
            self,
            *,
            src_fasta_file_path: str,
            dst_fastq_file_prefix: str,
            depth: int,
            llrg_executable_path: str,
            is_trusted: bool,
            strategy: PBSIM3_STRATEGY,
            hmm_method: str,
            hmm_model: str,
            samtools_executable_path: Optional[str],
            ccs_executable_path: Optional[str],
            ccs_num_threads: Optional[int],
            ccs_pass: int,
            preserve_intermediate_files: bool,
            other_args: List[str]
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :param is_trusted: Whether to skip input validation test.
        :param hmm_model: Name or Path of HMM Model.
        :param ccs_num_threads: [CCS] Number of threads used when invoking PBCCS.
        :param ccs_executable_path: [CCS] Path to PacBio PBCCS
        :param samtools_executable_path: [CCS] Path to Samtools.
        :param ccs_pass: Number of CCS passes, 1 for CLR, >1 for CCS.
        :param strategy: PBSIM3 stratergy, can be WGS or Isoform cDNA (trans).
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            preserve_intermediate_files=preserve_intermediate_files,
            is_trusted=is_trusted
        )

        if not is_trusted:
            validated_params = Pbsim3Adapter.validate_params(
                hmm_model=hmm_model,
                hmm_method=hmm_method,
                ccs_pass=ccs_pass,
                samtools_executable_path=samtools_executable_path,
                ccs_executable_path=ccs_executable_path
            )
            hmm_model = validated_params["hmm_model"]
        self._strategy = strategy
        self._samtools_executable_path = samtools_executable_path
        self._ccs_executable_path = ccs_executable_path
        self._ccs_pass = ccs_pass
        self._ccs_num_threads = ccs_num_threads

        possible_hmm_model_path = os.path.join(
            PBSIM3_DIST_DIR_PATH,
            f"{hmm_method.upper()}-{hmm_model}.model"
        )
        if os.path.exists(hmm_model):
            pass
        elif os.path.exists(possible_hmm_model_path):
            hmm_model = possible_hmm_model_path
        else:
            raise LLRGInitializationException(f"HMM Model {hmm_model} cannot be resolved!")

        if strategy not in PBSIM3_STRATEGY:
            raise LLRGInitializationException(f"strategy {strategy} should be in {PBSIM3_STRATEGY}!")

        if self._strategy == PBSIM3_STRATEGY.trans:
            self._input_path = os.path.join(
                self._tmp_dir, "transcript.tsv"
            )
            strategy_params = [
                "--transcript", self._input_path
            ]
        else:
            self._input_path = self._src_fasta_file_path
            strategy_params = [
                "--depth", str(depth),
                "--genome", self._input_path
            ]
        self._cmd = [
            llrg_executable_path,
            "--strategy", str(self._strategy),
            "--method", hmm_method,
            f"--{hmm_method}", hmm_model,
            "--prefix", os.path.join(self._tmp_dir, "tmp"),
            "--id-prefix", f"movie{uuid.uuid4()}",
            *strategy_params,
            "--pass-num", str(self._ccs_pass),
            *other_args
        ]

    def _pre_execution_hook(self) -> None:
        if self._strategy == PBSIM3_STRATEGY.trans:
            try:
                with get_writer(self._input_path) as transcript_writer, \
                        FastaViewFactory(
                            filename=self._src_fasta_file_path,
                            read_into_memory=True,
                            show_tqdm=False
                        ) as ff:
                    transcript_id = ff.chr_names[0]
                    sequence = ff.sequence(transcript_id)
                    transcript_writer.write("\t".join((
                        transcript_id,
                        str(self._depth),  # Forward
                        "0",  # Reverse
                        sequence
                    )) + "\n")
            except (KeyError, OSError, IndexError) as e:
                raise LLRGInitializationException(
                    f"Sequence {transcript_id} from file {self._src_fasta_file_path} failed!") from e

    def _ccs_to_fastq(self, prefix: str):
        subreads_bam_path = os.path.join(self._tmp_dir, f"{prefix}.subreads.bam")
        subreads_sam_path = os.path.join(self._tmp_dir, f"{prefix}.sam")
        subreads_xml_path = os.path.join(self._tmp_dir, f"{prefix}.subreads.xml")
        ccs_bam_path = os.path.join(self._tmp_dir, f"{prefix}.ccs.bam")
        ccs_xml_path = os.path.join(self._tmp_dir, f"{prefix}.ccs.xml")
        output_fastq_path = os.path.join(self._tmp_dir, f"{prefix}.ccs.fq")
        with get_writer(os.path.join(self._tmp_dir, "call_ccs.log"), is_binary=True) as log_writer:
            if self._exec_subprocess(
                    [
                        self._samtools_executable_path,
                        "view",
                        subreads_sam_path,
                        "-o", subreads_bam_path
                    ],
                    stdin=subprocess.DEVNULL,
                    stdout=log_writer,
                    stderr=log_writer
            ) != 0:
                return
            if jinja2 is not None:
                with get_writer(subreads_xml_path) as writer:
                    timestamp = time.localtime()
                    writer.write(PACB_SUBREAD_XML_TEMPLATE.render(
                        timestamp_file=time.strftime("%y-%m-%dT%H:%M:%S", timestamp),
                        timestamp_simple=time.strftime("%y%m%d_%H%m%S", timestamp),
                        bam_filepath=subreads_bam_path,
                        file_uuid=str(uuid.uuid4())
                    ))
            if self._exec_subprocess(
                    [
                        self._ccs_executable_path,
                        "--report-json", os.path.join(self._tmp_dir, f"{prefix}.ccs.report.json"),
                        "--report-file", os.path.join(self._tmp_dir, f"{prefix}.ccs.report.txt"),
                        "--log-level", "INFO",
                        "--log-file", os.path.join(self._tmp_dir, f"{prefix}.ccs.log"),
                        "--num-threads", str(self._ccs_num_threads),
                        subreads_xml_path if jinja2 is not None else subreads_bam_path,
                        ccs_xml_path
                    ],
                    stdin=subprocess.DEVNULL,
                    stdout=log_writer,
                    stderr=log_writer
            ) != 0:
                return
            with get_writer(output_fastq_path, is_binary=True) as writer:
                if self._exec_subprocess(
                        [
                            self._samtools_executable_path,
                            "fastq",
                            ccs_bam_path
                        ],
                        stdin=subprocess.DEVNULL,
                        stdout=writer,
                        stderr=log_writer
                ) != 0:
                    return

    def _post_execution_hook(self):
        if self._ccs_pass == 1:
            if self._strategy == PBSIM3_STRATEGY.wgs:
                automerge(
                    glob.glob(os.path.join(self._tmp_dir, "tmp_????.fastq")),
                    self._dst_fastq_file_prefix + ".fq"
                )
            else:
                autocopy(os.path.join(self._tmp_dir, "tmp.fastq"), self._dst_fastq_file_prefix + ".fq")
        else:
            if self._strategy == PBSIM3_STRATEGY.wgs:
                for fp in glob.glob(os.path.join(self._tmp_dir, "tmp_????.sam")):
                    prefix = os.path.basename(fp).split(".")[0]
                    self._ccs_to_fastq(prefix=prefix)
                automerge(
                    glob.glob(os.path.join(self._tmp_dir, "tmp_????.ccs.fq")),
                    self._dst_fastq_file_prefix + ".fq"
                )
            else:
                self._ccs_to_fastq(prefix="tmp")
                autocopy(os.path.join(self._tmp_dir, "tmp.ccs.fq"), self._dst_fastq_file_prefix + ".fq")

    @property
    def is_pair_end(self) -> bool:
        return False


def patch_frontend_parser(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser with ART arguments.
    """
    parser.add_argument(
        '-m',
        '--hmm_model',
        required=True,
        help="Basename of HMM file. "
             f"If you select errhmm in hmm_method, it would be {PBSIM3_ERRHMM_POSSIBLE_MODELS}"
             f"If you select qshmm in hmm_method, it would be {PBSIM3_QSHMM_POSSIBLE_MODELS}",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        "-M",
        "--hmm_method",
        required=True,
        help="Whether to simulate using quality score (as PBSIM2) or error profile (new)",
        nargs='?',
        type=str,
        action='store',
        choices=("errhmm", "qshmm")
    )
    parser.add_argument(
        "--ccs_pass",
        required=False,
        help="CCS Multipass Settings. Use 1 for CLR and others for CCS.",
        nargs='?',
        type=int,
        action='store',
        default=1
    )
    parser.add_argument(
        '--ccs_path',
        required=False,
        help="Executable name of ccs or pbccs. Omitted if ccs_pass == 1.",
        nargs='?',
        type=str,
        action='store',
        default="ccs"
    )
    parser.add_argument(
        '--samtools_path',
        required=False,
        help="Executable name of samtools. Omitted if ccs_pass == 1.",
        nargs='?',
        type=str,
        action='store',
        default="samtools"
    )
    parser.add_argument(
        '--strategy',
        required=False,
        help="Whether to use transcript (trans) mode or wgs (wgs) mode",
        choices=PBSIM3_STRATEGY,
        type=PBSIM3_STRATEGY.from_name,
        action='store',
        default="wgs"
    )
    parser = patch_frontend_argument_parser(parser, "--preserve_intermediate_files")
    return parser



"""
llrg.py -- Helper functions in Low-Level Read Generators.
"""

__all__ = (
    "patch_frontend_parser_tgs",
    "patch_frontend_parser_public",
    "patch_frontend_parser_bulk_rna_seq",
    "patch_frontend_parser_sc_rna_seq",
    "generate_callback",
    "BaseAssembler",
    "AssembleDumb",
    "AssemblePairEnd",
    "AssembleSingleEnd",
    "enhanced_which"
)

import argparse
import multiprocessing
import os
import shutil
import threading
import time
from abc import ABC, abstractmethod

from labw_utils.bioutils.datastructure.fasta_view import FastaViewFactory
from labw_utils.bioutils.parser.fastq import FastqWriter, FastqIterator
from labw_utils.bioutils.record.fastq import FastqRecord, MisFormattedFastqRecordError
from labw_utils.commonutils.lwio import file_system
from labw_utils.commonutils.lwio.safe_io import get_writer
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import Tuple, List, Callable, Optional
from yasim.helper.depth_io import DepthType

_lh = get_logger(__name__)


def remark_fastq_single_end(
        src_fastq_file_path: str,
        dst_fastq_file_writer: FastqWriter,
        transcript_id: str,
        transcript_depth: float,
        simulator_name: str,
        truncate_ratio_3p: float,
        truncate_ratio_5p: float
) -> Tuple[int, int]:
    """
    Re-mark all seq_id in FASTQ files with truncation.

    - R: ``{transcript_id}:{num_of_reads}:{transcript_depth}:{simulator_name}``

    :return : Tuple of (number of actually processed reads, number of actually processed bases)
    """
    num_of_reads = 0
    num_of_bases = 0
    for fastq_record in FastqIterator(src_fastq_file_path, show_tqdm=False):
        sequence, quality = fastq_record.sequence, fastq_record.quality
        seq_len = len(sequence)
        truncate_len_3p = int(seq_len * truncate_ratio_3p)
        truncate_len_5p = int(seq_len * truncate_ratio_5p)
        sequence = sequence[truncate_len_5p:seq_len - truncate_len_3p]
        quality = quality[truncate_len_5p:seq_len - truncate_len_3p]
        new_record = FastqRecord(
            seq_id=f"{transcript_id}:{num_of_reads}:{transcript_depth}:{simulator_name}",
            sequence=sequence,
            quality=quality
        )
        dst_fastq_file_writer.write(new_record)
        num_of_reads += 1
        num_of_bases += len(quality)
    return num_of_reads, num_of_bases


def remark_fastq_pair_end(
        src_fastq_1_file_path: str,
        src_fastq_2_file_path: str,
        dst_fastq_1_file_writer: FastqWriter,
        dst_fastq_2_file_writer: FastqWriter,
        transcript_id: str,
        transcript_depth: float,
        simulator_name: str
) -> Tuple[int, int]:
    """
    Re-mark all seq_id in FASTQ files, return number of reads

    - R1: ``{transcript_id}:{num_of_reads}:{transcript_depth}:{simulator_name}/1``
    - R2: ``{transcript_id}:{num_of_reads}:{transcript_depth}:{simulator_name}/2``

    :return : Tuple of (number of actually processed reads, number of actually processed bases)
    """
    num_of_reads = 0
    num_of_bases = 0
    for fastq_record_1, fastq_record_2 in zip(
            FastqIterator(src_fastq_1_file_path, show_tqdm=False),
            FastqIterator(src_fastq_2_file_path, show_tqdm=False)
    ):
        new_fastq_record_1 = FastqRecord(
            seq_id=f"{transcript_id}:{num_of_reads}:{transcript_depth}:{simulator_name}/1",
            sequence=fastq_record_1.sequence,
            quality=fastq_record_1.quality
        )
        new_fastq_record_2 = FastqRecord(
            seq_id=f"{transcript_id}:{num_of_reads}:{transcript_depth}:{simulator_name}/2",
            sequence=fastq_record_2.sequence,
            quality=fastq_record_2.quality
        )
        dst_fastq_1_file_writer.write(new_fastq_record_1)
        dst_fastq_2_file_writer.write(new_fastq_record_2)
        num_of_reads += 1
        num_of_bases += len(new_fastq_record_1)
        num_of_bases += len(new_fastq_record_2)
    return num_of_reads, num_of_bases


class AssembError(ValueError):
    ...


class AssemblerType(ABC, threading.Thread):
    """
    The Assembler that assembles sequenced transcripts of each isoform into one (SE) or two (PE) files.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

    @abstractmethod
    def run(self) -> None:
        """Execute the assembler."""
        raise NotImplementedError

    @abstractmethod
    def add_transcript_id(self, transcript_id: str) -> None:
        """
        Add transcript_id to assembler.

        :param transcript_id: Transcript ID of an finished LLRG run.
        """
        raise NotImplementedError

    @abstractmethod
    def terminate(self) -> None:
        """
        Mark that no further Transcript ID would be added.

        Assembler would terminate after processing all added Transcript IDs.
        """
        raise NotImplementedError


class BaseAssembler(AssemblerType, ABC):
    _transcript_ids_pending: List[str]
    _should_stop: bool
    _depth: DepthType
    _output_fastq_prefix: str
    _simulator_name: str
    _input_transcriptome_fasta_dir: str
    _input_fastq_dir: str

    def __init__(
            self,
            *,
            depth_data: DepthType,
            output_fastq_prefix: str,
            simulator_name: str,
            input_transcriptome_fasta_dir: str,
            input_fastq_dir: Optional[str] = None,
            **kwargs
    ):
        """
        Assembler Initialization.

        :param depth_data: Isoform-Level Depth Data.
        :param output_fastq_prefix: Prefix of output FASTQ used in LLRG orchestrators.
        :param simulator_name: Name of simulator. Used in FASTQ SeqID.
        :param input_transcriptome_fasta_dir: Directory of Transcriptome FASTA.
        :param kwargs: Other Miscellaneous arguments for compatibility. See subclass for details.
        """
        super().__init__()
        if input_fastq_dir is None:
            input_fastq_dir = output_fastq_prefix + ".d"
        self._transcript_ids_pending = []
        self._should_stop = False
        self._depth = depth_data
        self._output_fastq_prefix = output_fastq_prefix
        self._simulator_name = simulator_name
        self._input_transcriptome_fasta_dir = input_transcriptome_fasta_dir
        self._input_fastq_dir = input_fastq_dir

    def add_transcript_id(self, transcript_id: str):
        self._transcript_ids_pending.append(transcript_id)
        _lh.debug("ASSEMB: %s QUEUED", transcript_id)

    def terminate(self):
        _lh.info("ASSEMB: Termination signal received")
        self._should_stop = True

    @property
    def n_pending(self) -> int:
        return len(self._transcript_ids_pending)


class AssembleSingleEnd(BaseAssembler):
    """
    Assembler for single-end reads.
    """
    _truncate_ratio_3p: float
    _truncate_ratio_5p: float

    def __init__(
            self,
            *,
            depth_data: DepthType,
            output_fastq_prefix: str,
            simulator_name: str,
            input_transcriptome_fasta_dir: str,
            input_fastq_dir: Optional[str] = None,
            truncate_ratio_3p: float = 0,
            truncate_ratio_5p: float = 0,
            **kwargs
    ):
        """
        Assembler Initialization.

        :param depth_data: Isoform-Level Depth Data.
        :param output_fastq_prefix: Prefix of output FASTQ used in LLRG orchestrators.
        :param simulator_name: Name of simulator. Used in FASTQ SeqID.
        :param input_transcriptome_fasta_dir: Directory of Transcriptome FASTA.
        :param truncate_ratio_3p: Ratio of 3 prime truncation, range from 0 (no truncation) to 1 (truncate all).
        :param truncate_ratio_5p: Ratio of 5 prime truncation, range from 0 (no truncation) to 1 (truncate all).
        """
        super().__init__(
            depth_data=depth_data,
            output_fastq_prefix=output_fastq_prefix,
            simulator_name=simulator_name,
            input_transcriptome_fasta_dir=input_transcriptome_fasta_dir,
            input_fastq_dir=input_fastq_dir,
            **kwargs
        )
        self._truncate_ratio_3p = truncate_ratio_3p
        self._truncate_ratio_5p = truncate_ratio_5p

    def run(self):
        _lh.info("ASSEMB: START")
        with FastqWriter(self._output_fastq_prefix + ".fq") as writer, \
                get_writer(self._output_fastq_prefix + ".fq.stats") as stats_writer:
            stats_writer.write("\t".join((
                "TRANSCRIPT_ID",
                "INPUT_DEPTH",
                "SIMULATED_N_OF_READS",
                "SIMULATED_N_OF_BASES",
                "TRANSCRIBED_LENGTH",
                "SIMULATED_DEPTH"
            )) + "\n")

            def _assemb(transcript_id: str):
                _lh.debug("ASSEMB: %s START", transcript_id)
                this_fasta_path = os.path.join(self._input_transcriptome_fasta_dir, transcript_id + ".fa")
                this_fastq_basename = os.path.join(self._input_fastq_dir, transcript_id)
                this_fastq_path = this_fastq_basename + ".fq"
                if not file_system.file_exists(this_fastq_path):
                    _lh.warning("ASSEMB: %s: Skipped non-expressing transcript (FASTQ not exist)", transcript_id)
                    raise AssembError
                if not file_system.file_exists(this_fasta_path):
                    _lh.warning("ASSEMB: %s: Skipped none-existing transcript (FASTA not exist)", transcript_id)
                    raise AssembError
                try:
                    transcribed_length = FastaViewFactory(
                        filename=this_fasta_path,
                        read_into_memory=True,
                        show_tqdm=False
                    ).get_chr_length(transcript_id)
                except KeyError:
                    _lh.warning("ASSEMB: %s: Skipped none-existing transcript (FASTA Parsing Error)", transcript_id)
                    raise AssembError
                transcript_depth = self._depth[transcript_id]
                try:
                    num_of_reads, num_of_bases = remark_fastq_single_end(
                        src_fastq_file_path=this_fastq_path,
                        dst_fastq_file_writer=writer,
                        transcript_id=transcript_id,
                        transcript_depth=transcript_depth,
                        simulator_name=self._simulator_name,
                        truncate_ratio_3p=self._truncate_ratio_3p,
                        truncate_ratio_5p=self._truncate_ratio_5p
                    )
                except MisFormattedFastqRecordError:
                    _lh.warning("ASSEMB: %s: Skipped none-existing transcript (FASTQ Parsing Error)", transcript_id)
                    return
                stats_writer.write("\t".join((
                    transcript_id,  # "TRANSCRIPT_ID",
                    str(transcript_depth),  # "INPUT_DEPTH",
                    str(num_of_reads),  # "SIMULATED_N_OF_READS",
                    str(num_of_bases),  # "SIMULATED_N_OF_BASES",
                    str(transcribed_length),  # "TRANSCRIBED_LENGTH",
                    str(num_of_bases / transcribed_length)  # "SIMULATED_DEPTH"
                )) + "\n")
                _lh.debug("ASSEMB: %s FIN", transcript_id)
                stats_writer.flush()

            while not self._should_stop:
                if len(self._transcript_ids_pending) > 0:
                    try:
                        _assemb(self._transcript_ids_pending.pop())
                    except AssembError:
                        pass
                else:
                    time.sleep(0.01)
            _lh.info("ASSEMB: Post TERM -- Remaining %d transcripts", len(self._transcript_ids_pending))
            while len(self._transcript_ids_pending) > 0:
                try:
                    _assemb(self._transcript_ids_pending.pop())
                except AssembError:
                    pass
        _lh.info("ASSEMB: FIN")


class AssemblePairEnd(BaseAssembler):
    """
    Assembler for pair-end reads.
    """

    def run(self):
        def _assemb(transcript_id: str):
            _lh.debug("ASSEMB: %s START", transcript_id)
            this_fasta_path = os.path.join(self._input_transcriptome_fasta_dir, transcript_id + ".fa")
            this_fastq_basename = os.path.join(output_fastq_dir, transcript_id)
            this_fastq_r1_path = this_fastq_basename + "_1.fq"
            this_fastq_r2_path = this_fastq_basename + "_2.fq"
            if not file_system.file_exists(this_fastq_r1_path):
                _lh.warning("ASSEMB: %s: Skipped non-expressing transcript (FASTQ 1 not exist)", transcript_id)
                raise AssembError
            if not file_system.file_exists(this_fastq_r2_path):
                _lh.warning("ASSEMB: %s: Skipped non-expressing transcript (FASTQ 2 not exist)", transcript_id)
                raise AssembError
            if not file_system.file_exists(this_fasta_path):
                _lh.warning("ASSEMB: %s: Skipped none-existing transcript (FASTA not exist)", transcript_id)
                raise AssembError
            try:
                transcribed_length = FastaViewFactory(
                    filename=this_fasta_path,
                    read_into_memory=True,
                    show_tqdm=False
                ).get_chr_length(transcript_id)
            except KeyError:
                _lh.warning("ASSEMB: %s: Skipped none-existing transcript (FASTA Parsing Error)", transcript_id)
                raise AssembError
            transcript_depth = self._depth[transcript_id]
            try:
                num_of_reads, num_of_bases = remark_fastq_pair_end(
                    src_fastq_1_file_path=this_fastq_r1_path,
                    src_fastq_2_file_path=this_fastq_r2_path,
                    dst_fastq_1_file_writer=writer1,
                    dst_fastq_2_file_writer=writer2,
                    transcript_id=transcript_id,
                    transcript_depth=transcript_depth,
                    simulator_name=self._simulator_name
                )
            except MisFormattedFastqRecordError:
                _lh.warning("ASSEMB: %s: Skipped none-existing transcript (FASTQ Parsing Error)", transcript_id)
                return
            stats_writer.write("\t".join((
                transcript_id,  # "TRANSCRIPT_ID",
                str(transcript_depth),  # "INPUT_DEPTH",
                str(num_of_reads),  # "SIMULATED_N_OF_READS",
                str(num_of_bases),  # "SIMULATED_N_OF_BASES",
                str(transcribed_length),  # "TRANSCRIBED_LENGTH",
                str(num_of_bases / transcribed_length)  # "SIMULATED_DEPTH"
            )) + "\n")
            _lh.debug("ASSEMB: %s FIN", transcript_id)
            stats_writer.flush()

        _lh.info("ASSEMB: START")
        output_fastq_dir = self._output_fastq_prefix + ".d"
        with FastqWriter(self._output_fastq_prefix + "_1.fq") as writer1, \
                FastqWriter(self._output_fastq_prefix + "_2.fq") as writer2, \
                get_writer(self._output_fastq_prefix + ".fq.stats") as stats_writer:
            stats_writer.write("\t".join((
                "TRANSCRIPT_ID",
                "INPUT_DEPTH",
                "SIMULATED_N_OF_READS",
                "SIMULATED_N_OF_BASES",
                "TRANSCRIBED_LENGTH",
                "SIMULATED_DEPTH"
            )) + "\n")

            while not self._should_stop:
                if len(self._transcript_ids_pending) > 0:
                    try:
                        _assemb(self._transcript_ids_pending.pop())
                    except AssembError:
                        pass
                else:
                    time.sleep(0.01)
            _lh.info("ASSEMB: Post TERM -- Remaining %d transcripts", len(self._transcript_ids_pending))
            while len(self._transcript_ids_pending) > 0:
                try:
                    _assemb(self._transcript_ids_pending.pop())
                except AssembError:
                    pass

        _lh.info("ASSEMB: FIN")


class AssembleDumb(BaseAssembler):
    """Assembler that does not perform anything."""

    def run(self):
        _lh.info("ASSEMB: START")
        while not self._should_stop:
            time.sleep(0.01)
        _lh.info("ASSEMB: FIN")


def generate_callback(
        assembler: AssemblerType,
        transcript_id: str
) -> Callable[[threading.Thread], None]:
    """
    Generate callback function for assemblers.

    This function is used by schedulers.
    It would invoke :py:func:`AssemblerType.add_transcript_id` when an instance of LLRG is finished.

    :param assembler: Targeted assembler.
    :param transcript_id: Source Transcript ID.
    :return: Generated callback.
    """
    return lambda _: assembler.add_transcript_id(transcript_id)


def patch_frontend_parser_public(
        parser: argparse.ArgumentParser,
        llrg_name: str,
        default_llrg_executable_name: Optional[str] = None
) -> argparse.ArgumentParser:
    """
    Patch argument parser with commonly-used options.

    :param llrg_name: Name of LLRG.
    :param default_llrg_executable_name: Default LLRG Executable Name. None if is function-based.
    :param parser: Source parser.
    :return: Patched parser.
    """
    parser.add_argument(
        '-F',
        '--fastas',
        required=True,
        help="Directory of transcribed cDNA sequences in FASTA format from `transcribe` step",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-j',
        '--jobs',
        required=False,
        help="Number of LLRGs to be executed in parallel",
        nargs='?',
        type=int,
        action='store',
        default=multiprocessing.cpu_count()
    )
    parser.add_argument(
        '--simulator_name',
        required=False,
        help="Custom simulator name. Used in FASTQ tags\n"
             "This step is done in assemble, so if --not_perform_assemble is set, this option would be useless.",
        nargs='?',
        type=str,
        action='store',
        default=None
    )
    if default_llrg_executable_name is not None:
        parser.add_argument(
            '-e',
            '--llrg_executable_path',
            required=False,
            help=f"Executable name or absolute path of {llrg_name}",
            nargs='?',
            type=str,
            action='store',
            default=default_llrg_executable_name
        )
    return parser


def patch_frontend_parser_sc_rna_seq(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser for Single-Cell RNA-Seq

    :param parser: Source parser.
    :return: Patched parser.
    """
    parser.add_argument(
        '-d',
        '--depth',
        required=True,
        help="Path to Isoform-Level Depth directory generated by `generate_barcoded_isoform_replicates`",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-o',
        '--out',
        required=True,
        help="Path to output Directory.",
        nargs='?',
        type=str,
        action='store'
    )
    return parser


def patch_frontend_parser_bulk_rna_seq_assemble(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    return parser


def patch_frontend_parser_bulk_rna_seq(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser for Bulk-Cell RNA-Seq

    :param parser: Source parser.
    :return: Patched parser.
    """
    parser.add_argument(
        '-d',
        '--depth',
        required=True,
        help="Path to input Isoform-Level Depth TSV generated by `generate_depth_v2` or `generate_isoform_depth` step",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-o',
        '--out',
        required=True,
        help="Output transcript prefix. "
             "The output file would be {out}.fq for single-end "
             "and {out}_1.fq, {out}_2.fq for pair end.\n"
             "If --not_perform_assemble is set, would NOT generate FASTQ files but a {out}.d unassembled directory",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '--not_perform_assemble',
        help="Do NOT assemble the output of each isoforms into one file.",
        action='store_true'
    )
    return parser


def patch_frontend_parser_tgs(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """
    Patch argument parser with public option and
    ``--truncate_ratio_3p`` and ``--truncate_ratio_5p`` option.

    :param parser: Source parser.
    :return: Patched parser.
    """
    parser.add_argument(
        '--truncate_ratio_3p',
        required=False,
        help="[Single End TGS Only] Ratio of 3 prime truncation, range from 0 (no truncation) to 1 (truncate all).\n"
             "This step is done in assemble, so if --not_perform_assemble is set, this option would be useless.",
        nargs='?',
        type=float,
        action='store',
        default=0.0
    )
    parser.add_argument(
        '--truncate_ratio_5p',
        required=False,
        help="[Single End TGS Only] Ratio of 5 prime truncation, range from 0 (no truncation) to 1 (truncate all).\n"
             "This step is done in assemble, so if --not_perform_assemble is set, this option would be useless.",
        nargs='?',
        type=float,
        action='store',
        default=0.0
    )
    return parser


def enhanced_which(path_or_filename: str) -> str:
    """
    Enhanced :py:func:`shutil.which` method that would raise error on failure.

    :param path_or_filename: Path or filename you wish to resolve.
    :return: Absolute path found by :py:func:`shutil.which`.
    :raise FileNotFoundError: On failure.
    """
    resolved_path = shutil.which(path_or_filename)
    if resolved_path is None:
        raise FileNotFoundError(f"File {path_or_filename} not found!")
    else:
        path_or_filename = resolved_path
    return path_or_filename



"""
depth.py -- GEP Datastructure and Utils

Here contains generators for Gene Expression Profile (GEP).
"""

__all__ = (
    "GenerationFailureException",
    "simulate_depth_gmm_v2",
    "simulate_gene_level_depth_gmm",
    "simulate_isoform_variance_inside_a_gene",
    "generate_depth_replicates_uniform"
)

import functools
import math
from random import uniform

import numpy as np

from labw_utils.bioutils.datastructure.gene_view_v0_1_x.gene_view import GeneViewType
from labw_utils.commonutils.importer.tqdm_importer import tqdm
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.mlutils.ndarray_helper import describe
from labw_utils.typing_importer import List
from yasim.helper.depth_io import DepthType
from yasim.helper.gmm import GaussianMixture1D

_lh = get_logger(__name__)


class GenerationFailureException(RuntimeError):
    """Raised while generation of data was failed"""
    ...


def simulate_gene_level_depth_gmm(
        gv: GeneViewType,
        mu: float,
        low_cutoff: float,
        high_cutoff_ratio: float
) -> DepthType:
    """
    Simulate DGE using Gaussian mixture model. Used in YASIM 3.0
    The two ``cutoff`` parameters are for data generated by the model.
    After cutoff, the data would be scaled to make its mean equal to ``mu``.

    The data would be re-generated if it is all-zero or contains ``np.nan``.

    :param gv: GeneView of the GTF.
    :param mu: Targeted mean.
    :param low_cutoff: Depth lower than this value would be preserved to this value.
    :param high_cutoff_ratio: Depth higher than mu * high_cutoff_ratio would be preserved to this value.
    :return: Generated abundance.
    :raise GenerationFailureException: If the data was re-generated 20 times.
    """
    _lh.info("GEN GENE DEPTH: Loading GMM model...")
    gmm_model = GaussianMixture1D.import_model([
        # (0.14427447754677641, 1.5527662658235803, 0.5349602445403809),
        # (0.06838373058223621, 3.6990568545476674, 4.440892098500626e-16),
        # (0.3139879527820852, 2.3714543839859354, 0.3438604637707616),
        # (0.47335383908890194, 3.126802802022012, 0.3396120976402885)
        (0.1427655350146609, 1.5746800285632137, 0.5388586661428069),
        (0.06837661567650871, 3.699056854547668, 0.0),
        (0.31792430333258503, 2.369675925231799, 0.345735276372599),
        (0.4709335459762455, 3.1287227827640636, 0.33812788057291315)
    ])
    n_gene_ids = gv.number_of_genes
    depth = {}
    for i in range(20):
        _lh.info("GEN GENE DEPTH: Attempt %d: GMM...", i)
        data = np.power(10, gmm_model.rvs(size=2 * n_gene_ids) - 1) - 1
        _lh.info("GEN GENE DEPTH: Attempt %d: 1/4: Scaling...", i)
        data = data / np.mean(data) * mu  # Scale to similar mean; should have a ~10% error
        _lh.info("GEN GENE DEPTH: Attempt %d: 2/4: Filtering...", i)
        data = data[functools.reduce(
            np.logical_and,
            (
                data > 0,
                np.isfinite(data)
            )
        )]  # Filter data
        if len(data) < n_gene_ids:
            _lh.warning(
                "GEN GENE DEPTH: filtered data length (%d) smaller than required (%d); would regenerate",
                len(data), n_gene_ids
            )
            continue

        data = data[: n_gene_ids]
        _lh.info("GEN GENE DEPTH: Attempt %d: 3/4 Scaling...", i)
        data = data / np.mean(data) * mu  # Rescale to real mean
        _lh.info("GEN GENE DEPTH: Attempt %d: 4/4 Filtering...", i)
        data[mu * high_cutoff_ratio < data] = mu * high_cutoff_ratio
        data[data < low_cutoff] = low_cutoff
        break
    else:
        raise GenerationFailureException()
    _lh.info("GEN GENE DEPTH: Final distribution: %s", describe(data))
    for i, gene_id in enumerate(gv.iter_gene_ids()):
        depth[gene_id] = data[i]
    return depth


def simulate_isoform_variance_inside_a_gene(
        n: int,
        mu: float,
        alpha: int,
        low_cutoff: float,
        high_cutoff_ratio: float
) -> List[float]:
    """
    Generate isoform variance inside a gene using Zipf's Distribution

    :param n: Number of isoforms
    :param mu: Mean of expression data
    :param alpha: Zipf's Coefficient
    :param low_cutoff: Depth lower than this value would be preserved to this value.
    :param high_cutoff_ratio: Depth higher than mu * high_cutoff_ratio would be preserved to this value.
    :return: Generated abundance.
    :raise GenerationFailureException: If the data was re-generated 20 times.
    """
    if n == 1:
        return [mu]
    for _ in range(20):
        if alpha == 1:
            data = np.array([(rn ** (-1)) for rn in range(1, n + 1)])
        else:
            data = np.array([(alpha - 1) * (rn ** (-alpha)) for rn in range(1, n + 1)])
        if np.sum(np.isnan(data)) == 0 and np.sum(data) != 0:
            break
        else:
            _lh.warning("GEN ISOFORM DEPTH: NAN found in data; would regenerate")
    else:
        raise GenerationFailureException()
    data = data / np.mean(data) * mu
    data[data < low_cutoff] = low_cutoff
    data[data > mu * high_cutoff_ratio] = mu * high_cutoff_ratio
    np.random.shuffle(data)
    return data


def simulate_depth_gmm_v2(
        gv: GeneViewType,
        mu: float
) -> DepthType:
    """
    Simulate Isoform-level GEP using Gaussian mixture model using version 2 algorithm.

    :param gv: GeneView of the GTF.
    :param mu: Targeted mean.
    :return: Generated abundance.
    """
    gmm_model = GaussianMixture1D.import_model(
        [
            (0.2757332390925274, 1.115903346554058, 0.28730736063701984),
            (0.7242667609074726, 3.588018272411806, 1.5658127268321427)
        ]
    )
    transcript_ids = gv.iter_transcript_ids()
    n_transcript_ids = gv.number_of_transcripts
    depth = {}
    gmm_model.lintrans((math.log(mu) - 1) / gmm_model.positive_mean())
    data = np.exp(gmm_model.rvs(size=2 * n_transcript_ids) - 1)
    data = data[13 * mu >= data]
    data = data[data >= 0][:n_transcript_ids]
    i = 0
    for transcript_id in tqdm(iterable=transcript_ids, desc="Simulating..."):
        if data[i] != 0:
            depth[transcript_id] = data[i]
        i += 1
    return depth


def generate_depth_replicates_uniform(
        input_depth: DepthType,
        _range: float = 0.001
) -> DepthType:
    """
    Generate replications of depth in uniform distribution. Would be O = I + uniform(-_range, +_range) * I

    This is a naive method that modulates technical replicate.

    :param input_depth: Input depth.
    :param _range: Degree of variation to be added.
    :return: Mutated depth information.
    """
    return {
        k: v + uniform(-_range, +_range) * v
        for k, v in input_depth.items()
    }



"""
transcribe.py -- Proxy to ``sample_transcript`` in ``labw_utils.bioutils``.
"""
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List

try:
    from labw_utils.bioutils.main.sample_transcript import main as bc_main
except ImportError:
    from labw_utils.bioutils._main.sample_transcript import main as bc_main

try:
    from labw_utils.bioutils.main.transcribe import create_parser as create_parser
except ImportError:
    try:
        from labw_utils.bioutils._main.transcribe import create_parser as create_parser
    except ImportError:
        create_parser = None

_lh = get_logger(__name__)


def main(args: List[str]):
    _lh.warning(
        "DEPRECATION WARNING: "
        "`sample_transcript` at `yasim` had been deprecated. Use those in `labw_utils.bioutils` instead"
    )
    bc_main(args)



"""
assemble.py -- Assemble unassembled outputs.
"""

__all__ = (
    "main",
    "create_parser"
)

import argparse
import os
import time

from labw_utils.commonutils.stdlib_helper.argparse_helper import ArgumentParserWithEnhancedFormatHelp
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List, Type
from yasim.helper import llrg, depth_io
from yasim.helper.llrg import AssemblePairEnd, AssembleSingleEnd, AssemblerType

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(prog="python -m yasim pbsim", description=__doc__.splitlines()[1])
    parser.add_argument(
        '-F',
        '--fastas',
        required=True,
        help="Directory of transcribed cDNA sequences in FASTA format from `transcribe` step",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '--simulator_name',
        required=True,
        help="Custom simulator name. Used in FASTQ tags.",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-d',
        '--depth',
        required=True,
        help="Path to input Isoform-Level Depth TSV generated by `generate_depth_v2` or `generate_isoform_depth` step",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-o',
        '--out',
        required=True,
        help="Output transcript prefix. Should be prefix path to {out}.d directory.",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-i',
        '--input_fastq_dir',
        required=True,
        help="Input transcript prefix. Should be prefix path to {out}.d directory.",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '--is_pair_end',
        required=False,
        help="Whether to use Pair End (PE) Simulation",
        action='store_true'
    )
    parser = llrg.patch_frontend_parser_tgs(parser)
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    assembler_class: Type[AssemblerType]
    if args.is_pair_end:
        assembler_class = AssemblePairEnd
    else:
        assembler_class = AssembleSingleEnd
    depth_data = depth_io.read_depth(args.depth)
    assembler = assembler_class(
        depth_data=depth_data,
        output_fastq_prefix=os.path.abspath(args.out),
        input_fastq_dir=os.path.abspath(args.input_fastq_dir),
        simulator_name=args.simulator_name,
        input_transcriptome_fasta_dir=os.path.abspath(args.fastas),
        truncate_ratio_3p=args.truncate_ratio_3p,
        truncate_ratio_5p=args.truncate_ratio_5p
    )
    assembler.start()
    for transcript_id in depth_data.keys():
        assembler.add_transcript_id(transcript_id)

    assembler.terminate()
    while assembler.is_alive():
        _lh.info("ASSEMB: %s -- PENDING: %d", args.out, assembler.n_pending)
        time.sleep(1.0)
    assembler.join()



"""
extract_read_length_from_maf_gp.py -- Extraction of Read Length from MAF, General-Purposed

This script can be used to extract read length of all transcript ID from transcriptomically-aligned TGS RNA-Seq MAF
for assessing read completeness.

Synopsis: python -m yasim_scrips extract_quality_from_maf [MAF1] [[MAF2]...]

Arguments:
    [MAF1] [[MAF2]...] path to MAF files produced by PBSIM3 or LAST aligner.
"""
__all__ = (
    "main",
)

from labw_utils.typing_importer import List

from yasim_scripts.helper.maf_parser import maf_parse


def main(args: List[str]):
    for arg in args:
        print("\t".join(("ALIGNED_TRANSCRIPT_ID", "READ_LENGTH")))
        for maf_record in maf_parse(arg):
            aligned_transcript_id = maf_record[0]
            read_length = len(maf_record[3].replace("-", ""))
            print("\t".join((
                aligned_transcript_id,
                str(read_length)
            )))



"""
generate_depth_v2.py -- Generate Isoform-Level Depth using YASIM V2 API.
"""

__all__ = (
    "main",
    "create_parser"
)

import argparse

from labw_utils.bioutils.datastructure.gene_view_v0_1_x.gene_view import GeneViewFactory
from labw_utils.commonutils.stdlib_helper.argparse_helper import ArgumentParserWithEnhancedFormatHelp
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List
from yasim.helper import depth, depth_io
from yasim.helper.frontend import patch_frontend_argument_parser

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(prog="python -m yasim generate_depth_v2",
                                                  description=__doc__.splitlines()[1])
    parser = patch_frontend_argument_parser(parser, "-g")
    parser.add_argument(
        '-o', '--out',
        required=True,
        help="Path to output Isoform-Level depth TSV. Can be compressed.",
        nargs='?',
        type=str,
        action='store'
    )
    parser.add_argument(
        '-d', '--mu',
        required=False,
        help="Average depth",
        nargs='?',
        type=float,
        action='store',
        default=100
    )
    return parser


def main(args: List[str]):
    _lh.warning("DEPRECATION WARNING: "
                "The V2 API had been deprecated. Use `generate_gene_depth` and `generate_isoform_depth` instead")
    args = create_parser().parse_args(args)
    gv = GeneViewFactory.from_file(args.gtf)
    dge_data = depth.simulate_depth_gmm_v2(gv=gv, mu=args.mu)
    depth_io.write_depth(dge_data, args.out, feature_name="TRANSCRIPT_ID")



"""
dwgsim.py -- LLRG adapter for DWGSIM, a NGS DNA-Seq simulator
"""

__all__ = (
    "main",
    "create_parser"
)

import argparse

from labw_utils.commonutils.stdlib_helper.argparse_helper import ArgumentParserWithEnhancedFormatHelp
from labw_utils.typing_importer import List
from yasim.helper import llrg
from yasim.helper.rna_seq import bulk_rna_seq_frontend
from yasim.llrg_adapter import dwgsim


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(prog="python -m yasim dwgsim", description=__doc__.splitlines()[1])
    parser = llrg.patch_frontend_parser_public(
        parser,
        llrg_name="dwgsim",
        default_llrg_executable_name="dwgsim"
    )
    parser = llrg.patch_frontend_parser_bulk_rna_seq(parser)
    parser = dwgsim.patch_frontend_parser(parser)
    return parser


def main(args: List[str]):
    args, other_args = create_parser().parse_known_args(args)
    return bulk_rna_seq_frontend(
        transcriptome_fasta_dir=args.fastas,
        output_fastq_prefix=args.out,
        depth_file_path=args.depth,
        jobs=args.jobs,
        simulator_name="dwgsim" if args.simulator_name is None else args.simulator_name,
        adapter_args={
            "other_args": other_args,
            "preserve_intermediate_files": args.preserve_intermediate_files
        },
        assembler_args={},
        adapter_class=dwgsim.DwgsimAdapter,
        is_pair_end=True,
        llrg_executable_path=args.llrg_executable_path,
        not_perform_assemble=args.not_perform_assemble
    )



from labw_utils.commonutils.libfrontend import setup_frontend

from yasim import \
    __version__ as yasim_ver, \
    description

if __name__ == '__main__':
    setup_frontend(
        "yasim._main",
        description,
        yasim_ver,
    )



"""
generate_as_events.py -- Generate Alternative Splicing Events from Reference genome using YASIM V3 API.
"""

__all__ = (
    "main",
    "create_parser"
)

import argparse

from labw_utils.bioutils.datastructure.gene_view_v0_1_x.gene_view import GeneViewFactory
from labw_utils.commonutils.stdlib_helper.argparse_helper import ArgumentParserWithEnhancedFormatHelp
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List
from yasim.helper.as_events import ASManipulator
from yasim.helper.frontend import patch_frontend_argument_parser

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(prog="python -m yasim generate_as_events",
                                                  description=__doc__.splitlines()[1])
    parser = patch_frontend_argument_parser(parser, "-g")
    parser = patch_frontend_argument_parser(parser, "-f")
    parser.add_argument(
        '-c',
        '--complexity',
        required=True,
        help="Transcriptome Complexity Index, should be an integer between 1 and 9.",
        nargs='?',
        type=int,
        action='store'
    )
    parser.add_argument(
        '-o',
        '--out',
        required=True,
        help="Path to output genome annotation with Ground-Truth AS Events in GTF.",
        nargs='?',
        type=str,
        action='store'
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    gv = GeneViewFactory.from_file(args.gtf)
    _lh.info(
        "GEN AS EVENTS: Loaded %d genes with %d transcript",
        gv.number_of_genes,
        gv.number_of_transcripts
    )
    asm = ASManipulator(gv=gv)
    asm.run("ce", args.complexity)
    asm.to_file(args.out)



"""
art.py -- ART adapter for Illumina sequencing.
"""

__all__ = (
    "AVAILABLE_ILLUMINA_ART_SEQUENCER",
    "ArtAdapter",
    "patch_frontend_parser"
)

import argparse
import os

from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List, Tuple, Union, Final, Any, Mapping
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, autocopy, LLRGInitializationException

AVAILABLE_ILLUMINA_ART_SEQUENCER: Mapping[str, Tuple[str, List[int]]] = {
    "GA1": ("GenomeAnalyzer I", [36, 44]),
    "GA2": ("GenomeAnalyzer II", [50, 75]),
    "HS10": ("HiSeq 1000", [100]),
    "HS20": ("HiSeq 2000", [100]),
    "HS25": ("HiSeq 2500", [125, 150]),
    "HSXn": ("HiSeqX PCR free", [150]),
    "HSXt": ("HiSeqX TruSeq", [150]),
    "MinS": ("MiniSeq TruSeq", [50]),
    "MSv1": ("MiSeq v1", [250]),
    "MSv3": ("MSv3 - MiSeq v3", [250]),
    "NS50": ("NextSeq500 v2", [75])
}

_lh = get_logger(__name__)


class ArtAdapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper of ART.

    Cmdline Specs::

        cmd = [
            llrg_executable_path,
            "--fcov", str(self._depth),
            "--in", self._src_fasta_file_path,
            "--samout",
            "--noALN",
            "-l", rlen,
            "-O", os.path.join(self._tmp_dir, "tmp"),
            *other_args
        ]
    """

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    llrg_name: Final[str] = "art"
    _require_integer_depth: Final[bool] = False
    _capture_stdout: Final[bool] = False
    _is_pair_end: bool

    @staticmethod
    def validate_params(
            sequencer_name: str,
            read_length: int,
            pair_end_fragment_length_mean: int,
            pair_end_fragment_length_std: int,
            is_pair_end: bool,
            **kwargs
    ) -> Mapping[str, Any]:
        _ = kwargs
        retd = {}
        try:
            sequencer_profile = AVAILABLE_ILLUMINA_ART_SEQUENCER[sequencer_name]
        except KeyError as e:
            raise LLRGInitializationException(f"Sequencer Profile for sequencer {sequencer_name} not found!") from e
        if read_length not in sequencer_profile[1]:
            read_length_used = sequencer_profile[1][0]
            _lh.warning("ART: Read length %d not allowed, would use default %d.", read_length, read_length_used)
        else:
            read_length_used = read_length
        retd.update({"read_length": read_length_used})
        if is_pair_end:
            if pair_end_fragment_length_mean * pair_end_fragment_length_std == 0:
                raise LLRGInitializationException(
                    "Please set pair_end_fragment_length_mean and pair_end_fragment_length_std in PE simulation"
                )
        else:
            if pair_end_fragment_length_mean != 0 or pair_end_fragment_length_std != 0:
                _lh.warning(
                    "ART: PE Params pair_end_fragment_length_mean pair_end_fragment_length_std "
                    "should not be used in SE mode, ignored."
                )
        return retd

    def __init__(
            self,
            *,
            src_fasta_file_path: str,
            dst_fastq_file_prefix: str,
            depth: Union[int, float],
            llrg_executable_path: str,
            is_trusted: bool,
            sequencer_name: str,
            read_length: int,
            pair_end_fragment_length_mean: int,
            pair_end_fragment_length_std: int,
            is_pair_end: bool,
            preserve_intermediate_files: bool,
            other_args: List[str]
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param is_trusted: Whether to skip input validation test.
        :param sequencer_name: name of sequencer to simulate.
        :param read_length: Length of read. Should be defined in corresponding sequencer.
        :param pair_end_fragment_length_mean: [PE Only] The mean size of DNA/RNA fragments for paired-end simulations.
        :param pair_end_fragment_length_std: [PE Only] The standard deviation of DNA/RNA fragment size for
            paired-end simulations.
        :param is_pair_end: Whether the simulation should be pair end.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            is_trusted=is_trusted,
            preserve_intermediate_files=preserve_intermediate_files
        )

        if not is_trusted:
            validated_params = ArtAdapter.validate_params(
                sequencer_name=sequencer_name,
                read_length=read_length,
                pair_end_fragment_length_mean=pair_end_fragment_length_mean,
                pair_end_fragment_length_std=pair_end_fragment_length_std,
                is_pair_end=is_pair_end
            )
            read_length = validated_params["read_length"]
        self._is_pair_end = is_pair_end
        self._cmd = [
            self._llrg_executable_path,
            "--fcov", str(self._depth),
            "--in", self._src_fasta_file_path,
            "--samout",
            "--noALN",
            "-l", str(read_length),
            "--out", os.path.join(self._tmp_dir, "tmp"),
            *other_args
        ]
        if self._is_pair_end:
            self._cmd.extend([
                "-p",
                "--sdev", str(pair_end_fragment_length_std),
                "--mflen", str(pair_end_fragment_length_mean)
            ])

    def _post_execution_hook(self):
        if self._is_pair_end:
            autocopy(os.path.join(self._tmp_dir, "tmp1.fq"), self._dst_fastq_file_prefix + "_1.fq")
            autocopy(os.path.join(self._tmp_dir, "tmp2.fq"), self._dst_fastq_file_prefix + "_2.fq")
        else:
            autocopy(os.path.join(self._tmp_dir, "tmp.fq"), self._dst_fastq_file_prefix + ".fq")

    @property
    def is_pair_end(self) -> bool:
        return self._is_pair_end


def patch_frontend_parser(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser with ART arguments.
    """
    parser.add_argument(
        "--sequencer_name",
        required=False,
        help="Name of Illumina Sequencer to Simulate: " + ", ".join((
            f"{name} -- {AVAILABLE_ILLUMINA_ART_SEQUENCER[name][0]}"
            for name in AVAILABLE_ILLUMINA_ART_SEQUENCER.keys()
        )),
        nargs='?',
        choices=AVAILABLE_ILLUMINA_ART_SEQUENCER.keys(),
        type=str,
        action='store',
        default="HS25"
    )
    parser.add_argument(
        "--read_length",
        required=False,
        help="Read length. Sequencer -- Read Length Table: " + ", ".join((
            f"{v[0]} -- {v[1]}"
            for v in AVAILABLE_ILLUMINA_ART_SEQUENCER.values()
        )),
        nargs='?',
        type=int,
        action='store',
        default=0
    )
    parser.add_argument(
        '--pair_end_fragment_length_mean',
        required=False,
        help="[PE Only] The mean size of DNA/RNA fragments for paired-end simulations",
        nargs='?',
        type=int,
        action='store',
        default=0
    )
    parser.add_argument(
        '--pair_end_fragment_length_std',
        required=False,
        help="[PE Only] The standard deviation of DNA/RNA fragment size for paired-end simulations.",
        nargs='?',
        type=int,
        action='store',
        default=0
    )
    parser.add_argument(
        '--is_pair_end',
        required=False,
        help="Whether to use Pair End (PE) Simulation",
        action='store_true'
    )
    parser = patch_frontend_argument_parser(parser, "--preserve_intermediate_files")
    return parser



"""
dwgsim.py -- Wrapper of DWGSIM.
"""

__all__ = (
    "DwgsimAdapter",
)

import argparse
import os

from labw_utils.commonutils.lwio import file_system
from labw_utils.typing_importer import List, Union, Final, Mapping, Any
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, autocopy, NoOutputFileException


class DwgsimAdapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper of DWGSIM.

    Cmdline Specs::

        cmd = [
            llrg_executable_path,
            "-C", str(self._depth),
            *other_args,
            self._src_fasta_file_path,
            self._tmp_dir
        ]
    """
    llrg_name: Final[str] = "dwgsim"
    _require_integer_depth: Final[bool] = False
    _capture_stdout: Final[bool] = False

    @staticmethod
    def validate_params(**kwargs) -> Mapping[str, Any]:
        return {}

    def __init__(
            self,
            *,
            src_fasta_file_path: str,
            dst_fastq_file_prefix: str,
            depth: Union[int, float],
            llrg_executable_path: str,
            is_trusted: bool,
            preserve_intermediate_files: bool,
            other_args: List[str],
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :param is_trusted: Whether to skip input validation test.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            is_trusted=is_trusted,
            llrg_executable_path=llrg_executable_path,
            preserve_intermediate_files=preserve_intermediate_files
        )
        self._cmd = [
            llrg_executable_path,
            "-C", str(self._depth),
            *other_args,
            self._src_fasta_file_path,
            os.path.join(self._tmp_dir, "tmp")
        ]

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    def _post_execution_hook(self):
        try_read1_suffix = (
            ".bwa.read1.fastq.gz",
            ".bwa.read1.fastq",
            ".bwa.read1.fq.gz",
            ".bwa.read1.fq"
        )
        try_read2_suffix = (
            ".bwa.read2.fastq.gz",
            ".bwa.read2.fastq",
            ".bwa.read2.fq.gz",
            ".bwa.read2.fq"
        )
        for suffix_r1, suffix_r2 in zip(try_read1_suffix, try_read2_suffix):
            r1_file_path = os.path.join(self._tmp_dir, "tmp" + suffix_r1)
            r2_file_path = os.path.join(self._tmp_dir, "tmp" + suffix_r2)

            if not file_system.file_exists(r1_file_path) or \
                    not file_system.file_exists(r2_file_path):
                continue
            autocopy(r1_file_path, self._dst_fastq_file_prefix + "_1.fq")
            autocopy(r2_file_path, self._dst_fastq_file_prefix + "_2.fq")
            break
        else:
            raise NoOutputFileException("Unable to find output")

    @property
    def is_pair_end(self) -> bool:
        return True


def patch_frontend_parser(
        parser: argparse.ArgumentParser
) -> argparse.ArgumentParser:
    """
    Patch argument parser with DWGSIM arguments.
    """
    parser = patch_frontend_argument_parser(parser, "--preserve_intermediate_files")
    return parser



"""
plot_utils.py -- Utility functions in GMM fitting process.
"""

import numpy as np
import numpy.typing as npt

from labw_utils import UnmetDependenciesError
from labw_utils.mlutils.ndarray_helper import describe
from labw_utils.typing_importer import Optional, Any

try:

    import matplotlib
    from matplotlib import pyplot as plt
except ImportError as e:
    raise UnmetDependenciesError("matplotlib") from e

matplotlib.use('qtagg')

__all__ = ("plot",)


def plot(data: npt.NDArray, model: Optional[Any] = None, title: str = ""):
    """
    Plot how fitted GMM model performs over actual data
    """

    def normalize(_simulated_data: npt.NDArray) -> npt.NDArray:
        return _simulated_data[_simulated_data >= 0][:len(data)]

    data = np.array(data)
    n_bins = np.arange(0, 10, 0.1)
    if model is not None:
        print(str(model))

        simulated_data = model.rvs(2 * len(data))
        simulated_data_n = normalize(simulated_data)

        fig, axes = plt.subplots(1, 1)
        axes.hist(data, bins=n_bins, density=True, color="blue", alpha=0.2)
        axes.hist(simulated_data_n, bins=n_bins, density=True, color="red", alpha=0.2)
        print("data:                        " + describe(data))
        print("simulated_data:              " + describe(simulated_data))
        print("simulated_data (normalized): " + describe(simulated_data_n))
        plt.title(f"{title} Red: Simulated, Blue: Real")
        plt.show()
    else:
        plt.hist(data, bins=n_bins, density=True, color="blue", alpha=0.2)
        plt.title(f"{title} Blue: Real")
        plt.show()



"""
as_events.py -- Generate AS Events

This file contains necessary modules in *de novo* generation of AS events.
It would reshape Number of Isoforms per Gene on Reference Genome in a log-normal distribution,
with new isoforms introduced by creating AS events and redundant isoforms removed.
"""
__all__ = (
    "ASManipulator",
)

import random

import numpy as np
from scipy.stats import lognorm

from labw_utils.bioutils.datastructure.gene_view_v0_1_x.gene_view import GeneViewType
from labw_utils.bioutils.datastructure.gene_view_v0_1_x.gv_feature_proxy import Gene
from labw_utils.commonutils.importer.tqdm_importer import tqdm
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List, Callable, Iterable

ORGANISM_PARAMS = {
    "ce": (1.0274021145895147, 0.6524307003952217, 0.41902707818534024)
}
"""
(s, loc, scale) of a log-normal distribution.
"""

ORGANISM_WEIGHTS = {
    "ce": (
        11714,
        5886,
        8222,
        4371
    )
}
"""
Weight of each different AS event type.
"""

_lh = get_logger(__name__)


class ImpossibleToGenerateASEventError(ValueError):
    ...


class ASManipulator:
    _gv: GeneViewType

    def __init__(self, gv: GeneViewType):
        """
        Creation of an AS manipulator.

        :param gv: GeneView.

        .. warning::
            The Gene View passed would be internally modified!
        """
        self._gv = gv
        for transcript_id in self._gv.iter_transcript_ids():
            self._gv.transcript_sort_exons(transcript_id, "none")

    def _generate_one_as_event(
            self,
            core_func: Callable[[str], str],
            gene: Gene,
            max_try: int
    ):
        for _ in range(max_try):
            transcript_id = random.choice(list(gene.iter_transcript_ids()))
            try:
                new_transcript_id = core_func(transcript_id)
            except IndexError:
                # lh.debug(f"{core_func.__name__} IndexError")
                continue
            self._gv.transcript_sort_exons(new_transcript_id, "none")
            duplicated_transcript_id = gene.check_whether_one_transcript_duplicates_with_others(new_transcript_id)
            if duplicated_transcript_id is not None:
                self._gv.del_transcript(new_transcript_id)
                # lh.debug(f"{core_func.__name__} Duplicated")
            elif gene.get_transcript(new_transcript_id).transcribed_length < 250:
                self._gv.del_transcript(new_transcript_id)
                # lh.debug(f"{core_func.__name__} Too short")
            else:
                return
        _lh.debug("GEN AS EVENT: Gene %s: %d trials exhausted", gene.gene_id, max_try)
        raise ImpossibleToGenerateASEventError

    def _generate_multiple_as_events(
            self,
            core_funcs: Iterable[Callable[[str], str]],
            gene: Gene,
            max_try: int = 100
    ):
        for core_func in core_funcs:
            self._generate_one_as_event(core_func=core_func, gene=gene, max_try=max_try)

    def _core_perform_exon_skipping(self, transcript_id: str) -> str:
        number_of_exons = self._gv.get_transcript(transcript_id).number_of_exons
        # set the percent of knocked out exons
        percent = random.random() * 0.3 + 0.3
        # get the number of exons n to be knocked out by multiplying total exon number of transcript
        new_transcript_id = self._gv.duplicate_transcript(transcript_id)
        es_ids = random.sample(range(0, number_of_exons), int(number_of_exons * percent))
        if len(es_ids) == 0:
            raise ImpossibleToGenerateASEventError
        _lh.debug(f"GEN AS EVENT: {new_transcript_id}: ES {es_ids}, total={number_of_exons}")
        for exon_id in es_ids:
            self._gv.del_exon(new_transcript_id, exon_id)
        return new_transcript_id

    def _core_perform_intron_retention(self, transcript_id: str) -> str:
        number_of_exons = self._gv.get_transcript(transcript_id).number_of_exons
        start_exon_id = random.choice(range(0, number_of_exons - 2))
        stop_exon_id = start_exon_id + 1
        new_transcript_id = self._gv.duplicate_transcript(transcript_id)
        _lh.debug(f"GEN AS EVENT: {new_transcript_id}: IR {start_exon_id}-{stop_exon_id}, total={number_of_exons}")
        new_transcript = self._gv.get_transcript(new_transcript_id)
        new_transcript.get_nth_exon(start_exon_id).end = new_transcript.get_nth_exon(stop_exon_id).end
        self._gv.del_exon(new_transcript_id, stop_exon_id)
        return new_transcript_id

    def _core_perform_alternative_3p_splicing(self, transcript_id: str) -> str:
        transcript = self._gv.get_transcript(transcript_id)
        # randomly pick an exon for splicing
        exon_id = random.randint(0, (transcript.number_of_exons - 1))
        # randomly generate the percent to shorten
        exon_len = transcript.get_nth_exon(exon_id).transcribed_length
        intron_len = transcript.get_intron_length(exon_id)
        splice_perc = (random.random() - 0.5) * 1.6
        delta = int(min(exon_len, intron_len) * splice_perc)
        new_transcript_id = self._gv.duplicate_transcript(transcript_id)
        _lh.debug(
            f"GEN AS EVENT: {new_transcript_id}: A3P {exon_id} (EXON_LEN={exon_len}, INTRON_LEN={intron_len}), {delta}")
        new_transcript = self._gv.get_transcript(new_transcript_id)
        new_transcript.get_nth_exon(exon_id).end += delta
        return new_transcript_id

    def _core_perform_alternative_5p_splicing(self, transcript_id: str) -> str:
        transcript = self._gv.get_transcript(transcript_id)
        # randomly pick an exon for splicing
        exon_id = random.randint(0, transcript.number_of_exons - 1)
        # randomly generate the percent to shorten
        exon_len = transcript.get_nth_exon(exon_id).transcribed_length
        intron_len = transcript.get_intron_length(exon_id - 1)
        splice_perc = (random.random() - 0.5) * 1.6
        delta = int(min(exon_len, intron_len) * splice_perc)
        new_transcript_id = self._gv.duplicate_transcript(transcript_id)
        _lh.debug(
            f"GEN AS EVENT: {new_transcript_id}: A5P {exon_id} (EXON_LEN={exon_len}, INTRON_LEN={intron_len}), {delta}")
        new_transcript = self._gv.get_transcript(new_transcript_id)
        new_transcript.get_nth_exon(exon_id).start += delta
        return new_transcript_id

    def _perform_alternative_splicing(self, gene: Gene, organism: str) -> None:
        core_funcs: Iterable[Callable[[str], str]] = random.choices(
            population=(
                (self._core_perform_alternative_3p_splicing,),
                (self._core_perform_alternative_5p_splicing,),
                (self._core_perform_exon_skipping,),
                (self._core_perform_intron_retention,),
            ),
            weights=ORGANISM_WEIGHTS[organism],
            k=1
        )[0]
        self._generate_multiple_as_events(core_funcs, gene)

    def _try_generate_n_isoform_for_a_gene(self, gene: Gene, n: int, organism: str):
        transcript_ids_to_del = []
        for transcript in gene.iter_transcripts():
            if transcript.transcribed_length < 250:
                transcript_ids_to_del.append(transcript.transcript_id)
        for transcript_id in transcript_ids_to_del:
            self._gv.del_transcript(transcript_id, auto_remove_empty_gene=False)
        if gene.number_of_transcripts == 0:
            raise ImpossibleToGenerateASEventError("Gene cannot transcribed to length >= 250")
        elif gene.number_of_transcripts == n:
            return
        elif gene.number_of_transcripts > n:
            gek = list(gene.iter_transcript_ids())
            while gene.number_of_transcripts > n:
                self._gv.del_transcript(gek.pop())
        elif gene.number_of_transcripts < n:
            number_of_fail = 0
            while True:
                try:
                    self._perform_alternative_splicing(gene, organism)
                except ImpossibleToGenerateASEventError:
                    number_of_fail += 1
                if number_of_fail > 2 * n:
                    raise ImpossibleToGenerateASEventError("Generation FAILED!")
                elif gene.number_of_transcripts == n:
                    break

    def run(self, organism: str, compl_idx: int) -> None:
        """
        Execute the ASManipulator.

        :param organism: Name of the organism. Allowed are:

            - ``ce``: *C. Elegans*.
        :param compl_idx: The Transcriptome Complexity Index. Larger for larger Number of Isoforms in a Gene.
        """
        _lh.info("GEN AS EVENT: START")
        if organism not in ORGANISM_PARAMS.keys():
            raise ValueError(f"no organism {organism} available! Available: {ORGANISM_PARAMS.keys()}")
        fit_tuple = ORGANISM_PARAMS[organism]
        gene_ids_to_del: List[str] = []
        all_gene_ids = list(self._gv.iter_gene_ids())
        targeted_nipg = np.array(
            lognorm.rvs(
                fit_tuple[0],
                loc=fit_tuple[1],
                scale=fit_tuple[2],
                size=len(all_gene_ids) * 5
            ),
            dtype=float
        )
        minvalue = np.min(targeted_nipg)
        targeted_nipg = (targeted_nipg - minvalue) * compl_idx / np.mean(targeted_nipg) + minvalue
        targeted_nipg = targeted_nipg[np.logical_and(1 < targeted_nipg, targeted_nipg < 25)][:len(all_gene_ids)]
        targeted_nipg = np.array(np.sort(targeted_nipg), dtype=int)
        reference_nipg = [
            self._gv.get_gene(gene_id).number_of_transcripts for gene_id in all_gene_ids
        ]
        reference_nipg_rank = np.argsort(reference_nipg)
        for gene_index in tqdm(
                iterable=range(len(all_gene_ids)),
                desc="Generating isoforms...",
                total=self._gv.number_of_genes
        ):
            gene = self._gv.get_gene(all_gene_ids[gene_index])
            n = targeted_nipg[reference_nipg_rank[gene_index]]
            try:
                self._try_generate_n_isoform_for_a_gene(gene, n, organism)
            except ImpossibleToGenerateASEventError:
                gene_ids_to_del.append(gene.gene_id)

        _lh.info(
            "GEN AS EVENT: Will remove %d genes out of %d",
            len(gene_ids_to_del),
            self._gv.number_of_genes

        )
        for gene_id in gene_ids_to_del:
            self._gv.del_gene(gene_id)
        self._gv.standardize()
        _lh.info("GEN AS EVENT: FIN")

    def to_file(self, dst_gtf_file_path: str) -> None:
        """
        Write the internal GeneView to GTF.

        :param dst_gtf_file_path: Path to GTF to be written.
        """
        self._gv.to_file(dst_gtf_file_path)



"""
frontend.py -- Utilities for other ``main`` functions.
"""

__all__ = (
    "patch_frontend_argument_parser",
)

import argparse

from labw_utils.bioutils.comm_frontend_opts import FrontendOptSpecs, FrontendOptSpec

FrontendOptSpecs.add(FrontendOptSpec(
    '-d', '--depth',
    required=True,
    help="Path to input Isoform-Level Depth TSV. Can be compressed.",
    nargs='?',
    type=str,
    action='store'
))

FrontendOptSpecs.add(FrontendOptSpec(
    '-b', '--barcodes',
    required=True,
    help="Path to input barcode TXT. Can be compressed.",
    nargs='?',
    type=str,
    action='store'
))
FrontendOptSpecs.add(FrontendOptSpec(
    '--low_cutoff',
    required=False,
    help="Depth lower than this value would be this value.",
    nargs='?',
    type=float,
    action='store',
    default=0.01
))
FrontendOptSpecs.add(FrontendOptSpec(
    '--high_cutoff_ratio',
    required=False,
    help="Depth higher than `mu * high_cutoff_ratio` would be `mu * high_cutoff_ratio`",
    nargs='?',
    type=float,
    action='store',
    default=200
))
FrontendOptSpecs.add(FrontendOptSpec(
    '--preserve_intermediate_files',
    required=False,
    help="Do not remove intermediate files.",
    action='store_true',
))


def patch_frontend_argument_parser(parser: argparse.ArgumentParser, argname: str) -> argparse.ArgumentParser:
    return FrontendOptSpecs.patch(parser, argname)



from labw_utils.commonutils.libfrontend import setup_frontend

from yasim import __version__
from yasim_scripts import description

if __name__ == '__main__':
    setup_frontend(
        "yasim_scripts._main",
        description,
        __version__
    )



"""
generate_isoform_replicates.py -- Generate Technical Replicates using YASIM V3 API.
"""

__all__ = (
    "create_parser",
    "main"
)

import argparse

from labw_utils.commonutils.stdlib_helper.argparse_helper import ArgumentParserWithEnhancedFormatHelp
from labw_utils.typing_importer import List
from yasim.helper import depth, depth_io
from yasim.helper.frontend import patch_frontend_argument_parser


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_isoform_replicates",
        description=__doc__.splitlines()[1]
    )
    parser = patch_frontend_argument_parser(parser, "-d")
    parser.add_argument(
        '-n',
        '--num_replicates',
        required=False,
        help="Number of Replicates to be generated",
        nargs='?',
        type=int,
        action='store',
        default=3
    )
    parser.add_argument(
        '-r',
        '--range',
        required=False,
        help="Range of Generated Data",
        nargs='?',
        type=float,
        action='store',
        default=0.1
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    depth_data = depth_io.read_depth(args.depth)
    for i in range(args.num_replicates):
        depth_io.write_depth(depth.generate_depth_replicates_uniform(
            depth_data,
            args.range
        ), f"{args.depth}.{i}", "TRANSCRIPT_ID")
